<?php declare(strict_types=1);
namespace Grifart\AssertFunction;

final class FunctionSignatureAssertionError extends \AssertionError {

	const PATH_DEPTH = 2;

	/** @internal use named constructors instead */
	public function __construct(\ReflectionFunction $reflection, string $message, \Throwable $previous = NULL)
	{
		parent::__construct(
			self::location($reflection) . ' ' . $message,
			0,
			$previous
		);
	}

	private static function location(\ReflectionFunction $reflection): string
	{
		$startLine = (string) $reflection->getStartLine();
		$endLine   = (string) $reflection->getEndLine();

		$linePosition = ($startLine === $endLine) ? $startLine
			: ($startLine . '-' . $endLine);
		return self::shortenPath($reflection->getFileName()) . ':' . $linePosition;
	}


	public static function wrongNumberOrArguments(\ReflectionFunction $reflection, int $numberOfParameters, int $actualCount): self
	{
		return new self($reflection,  "Wrong number of arguments in given function. Expected $numberOfParameters, $actualCount given.");
	}


	public static function missingReturnType(\ReflectionFunction $reflection, string $type): self
	{
		return new self($reflection, "Function is required to have return type of type $type.");
	}


	private static function shortenPath(string $path): string
	{
		$pathArr = explode(DIRECTORY_SEPARATOR, $path);
		$pathNumParts = count($pathArr);
		$pathParts = array_filter($pathArr, function ($val, $idx) use ($pathNumParts) { // only last x parts
			return $idx >= $pathNumParts - self::PATH_DEPTH;
		}, ARRAY_FILTER_USE_BOTH);

		return implode('/', $pathParts); // normalized shortened readable path
	}


	public static function parameterNullability(\ReflectionFunction $functionReflection, \ReflectionParameter $param, bool $expectedNullable, bool $actuallyNullable): self
	{
		$paramIndex = $param->getPosition() + 1; // 0-based indexes are not human-readable
		return new self(
			$functionReflection,
			$expectedNullable
				? "Parameter #{$paramIndex} (\${$param->getName()}) is expected to be nullable. But it is required."
				: "Parameter #{$paramIndex} (\${$param->getName()}) is expected to be required. But it is nullable."
		);

	}


	public static function wrongParameterType(\ReflectionFunction $functionReflection, \ReflectionParameter $parameterReflection, string $expectedParameterType, \ReflectionType $actualParameterType): self
	{
		$paramIndex = ($parameterReflection->getPosition() + 1);
		return new self(
			$functionReflection,
			"Given function parameter #{$paramIndex} \${$parameterReflection->getName()} has type {$parameterReflection->getType()}. However this parameter is expected to be {$expectedParameterType}."
		);
	}


	public static function wrongReturnTypeNullability(\ReflectionFunction $reflectionFunction, bool $expectingNullable, bool $actuallyNullable): self
	{
		return new self(
			$reflectionFunction,
			$expectingNullable
				? 'Return-type is expected to be nullable. But it is required.'
				: 'Return-type is expected to be required. But it is nullable.'
		);
	}


	public static function wrongReturnType(\ReflectionFunction $reflection, string $expectedReturnType, string $actualReturnType): self
	{
		return new self(
			$reflection,
			"Expected return type of type '$expectedReturnType', but given function declares '$actualReturnType'."
		);
	}
};
