<?php declare(strict_types=1);
/**
 * This file is part of assert-function-signature.
 */

namespace Grifart\AssertFunction;

final class SignatureAssertionUtil
{

	public function checkSignature(callable $function, array $parameters, ?string $expectedReturnType): void
	{
		$reflection = new \ReflectionFunction($function);

		// NUMBER OF PARAMETERS CHECK:
		$numberOfParameters = $reflection->getNumberOfParameters();
		if($numberOfParameters !== count($parameters)) {
			throw FunctionSignatureAssertionError::wrongNumberOrArguments($reflection, count($parameters), $numberOfParameters);
		}

		// PARAMETER TYPES CHECK:
		$i = 0;
		/** @var string[] $parameters */
		foreach($parameters AS $parameter) {
			assert(is_string($parameter));

			$parameterReflection = $reflection->getParameters()[$i++];
			assert($parameterReflection instanceof \ReflectionParameter);

			$this->checkFunctionParameter(
				$reflection,
				$parameterReflection,
				...$this->parseType($parameter)
			);
		}


		// RETURN TYPE:
		if($expectedReturnType !== NULL) {
			if(!$reflection->hasReturnType()) {
				throw FunctionSignatureAssertionError::missingReturnType($reflection, $expectedReturnType);
			}

			if ($reflection->hasReturnType()) {
				list($expectedReturnType, $expectingNullable) = $this->parseType($expectedReturnType);
				$returnTypeReflection = $reflection->getReturnType();

				$actuallyNullable = $returnTypeReflection->allowsNull();
				if ($expectingNullable !== $actuallyNullable) {
					throw FunctionSignatureAssertionError::wrongReturnTypeNullability(
						$reflection,
						$expectingNullable,
						$actuallyNullable
					);
				}
				if ($expectedReturnType !== (string) $reflection->getReturnType()) {
					throw FunctionSignatureAssertionError::wrongReturnType(
						$reflection,
						$expectedReturnType,
						(string) $reflection->getReturnType()
					);
				}
			}
		}
	}

	private function parseType(string $type): array
	{
		$cleanType = $type; // e.g. ?Namespace\Class
		$nullable = FALSE;
		if ($type[0] === '?') {
			$nullable = TRUE;
			$cleanType = substr($type, 1); // without leading '?'
		}
		return [$cleanType, $nullable];
	}

	private function checkFunctionParameter(
		\ReflectionFunction $functionReflection,
		\ReflectionParameter $parameterReflection,
		string $expectedParameterType,
		bool $expectedNullable
	): void
	{
		$reflectionType = $parameterReflection->getType();
		assert($reflectionType instanceof \ReflectionType);

		// checks:
		$actuallyNullable = $reflectionType->allowsNull();
		if ($expectedNullable !== $actuallyNullable) {
			throw FunctionSignatureAssertionError::parameterNullability(
				$functionReflection,
				$parameterReflection,
				$expectedNullable,
				$actuallyNullable
			);
		}
		if ($expectedParameterType !== (string) $reflectionType) {
			throw FunctionSignatureAssertionError::wrongParameterType(
				$functionReflection,
				$parameterReflection,
				$expectedParameterType,
				$reflectionType
			);
		}
	}

}
